use cgmath::{Deg, EuclideanSpace, Matrix2, Point2, Point3, Vector2};

use crate::{
	content,
	physics::PhysicsBody,
	render::{Sprite, SpriteTexture, Spriteable, SubSprite},
};

use super::{game::Projectile, InputStatus};

pub struct ShipControls {
	pub left: bool,
	pub right: bool,
	pub thrust: bool,
	pub guns: bool,
}

impl ShipControls {
	pub fn new() -> Self {
		ShipControls {
			left: false,
			right: false,
			thrust: false,
			guns: false,
		}
	}
}

pub struct ShipTickResult {
	pub projectiles: Vec<Projectile>,
}

pub struct Ship {
	pub physicsbody: PhysicsBody,
	pub controls: ShipControls,

	sprite: SpriteTexture,
	size: f32,
	engines: Vec<content::Engine>,
	guns: Vec<content::ShipGun>,
}

impl Ship {
	pub fn new(ct: &content::Ship, pos: Point2<f32>) -> Self {
		Ship {
			physicsbody: PhysicsBody::new(pos),
			sprite: SpriteTexture(ct.sprite.clone()),
			size: ct.size,
			engines: ct.engines.clone(),
			guns: ct.guns.clone(),
			controls: ShipControls::new(),
		}
	}

	pub fn update_controls(&mut self, input: &InputStatus) {
		self.controls.thrust = input.key_thrust;
		self.controls.right = input.key_right;
		self.controls.left = input.key_left;
		self.controls.guns = input.key_guns;
	}

	pub fn fire_guns(&mut self) -> Vec<Projectile> {
		let mut out = Vec::new();
		for i in &mut self.guns {
			if i.active_cooldown > 0.0 {
				continue;
			}
			i.active_cooldown = i.cooldown;

			let p = self.physicsbody.pos
				+ (Matrix2::from_angle(self.physicsbody.angle) * i.pos.to_vec());

			out.push(Projectile {
				position: p,
				velocity: self.physicsbody.vel
					+ (Matrix2::from_angle(self.physicsbody.angle) * Vector2 { x: 0.0, y: 400.0 }),
				angle: self.physicsbody.angle,
				sprite: SpriteTexture("projectile::blaster".into()),
				lifetime: 5.0,
			})
		}
		return out;
	}

	pub fn tick(&mut self, t: f32) -> ShipTickResult {
		if self.controls.thrust {
			self.physicsbody.thrust(50.0 * t);
		}

		if self.controls.right {
			self.physicsbody.rot(Deg(35.0) * t);
		}

		if self.controls.left {
			self.physicsbody.rot(Deg(-35.0) * t);
		}

		let p = if self.controls.guns {
			self.fire_guns()
		} else {
			Vec::new()
		};

		self.physicsbody.tick(t);
		for i in &mut self.guns {
			i.active_cooldown -= t;
		}

		return ShipTickResult { projectiles: p };
	}
}

impl Spriteable for Ship {
	fn get_sprite(&self) -> Sprite {
		let engines = if self.controls.thrust {
			Some(
				self.engines
					.iter()
					.map(|e| SubSprite {
						pos: Point3 {
							x: e.pos.x,
							y: e.pos.y,
							z: 1.0,
						},
						texture: SpriteTexture("flare::ion".to_owned()),
						angle: Deg(0.0),
						size: e.size,
					})
					.collect(),
			)
		} else {
			None
		};

		Sprite {
			pos: (self.physicsbody.pos.x, self.physicsbody.pos.y, 1.0).into(),
			texture: self.sprite.clone(), // TODO: sprite texture should be easy to clone
			angle: self.physicsbody.angle,
			size: self.size,
			children: engines,
		}
	}
}
